/* hgRefSeqMrna - Load refSeq mRNA alignments and other info into genieKnown table. */
#include "common.h"
#include "linefile.h"
#include "hash.h"
#include "cheapcgi.h"
#include "psl.h"
#include "jksql.h"
#include "hdb.h"
#include "hgRelate.h"
#include "obscure.h"

/* Variables that can be set from command line. */
boolean clTest = FALSE;
boolean clDots = 10;

void usage()
/* Explain usage and exit. */
{
errAbort(
  "hgRefSeqMrna - Load refSeq mRNA alignments and other info into genieKnown table\n"
  "usage:\n"
  "   hgRefSeqMrna database refSeq.ra refSeq.psl loc2ref\n"
  "options:\n"
  "   -dots=N Output a dot every N records parsed\n"
  "   -test Generates tab files but doesn't actually load database.\n");
}

struct hash *loadNameTable(struct sqlConnection *conn, 
    char *tableName, int hashSize)
/* Create a hash and load it up from table. */
{
char query[128];
struct sqlResult *sr;
char **row;
struct hash *hash = newHash(hashSize);

sprintf(query, "select id,name from %s", tableName);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    hashAdd(hash, row[1], intToPt(sqlUnsigned(row[0])));
    }
sqlFreeResult(&sr);
return hash;
}

int putInNameTable(struct hash *hash, FILE *f, char *name)
/* Add to name table if it isn't there already. 
 * Return ID of name in table. */
{
struct hashEl *hel;
if (name == NULL)
    return 0;
if ((hel = hashLookup(hash, name)) != NULL)
    return ptToInt(hel->val);
else
    {
    int id = hgNextId();
    fprintf(f, "%u\t%s\n", id, name);
    hashAdd(hash, name, intToPt(id));
    return id;
    }
}

struct refSeqInfo
/* Information on one refSeq. */
    {
    struct refSeqInfo *next;
    char *mrnaAcc;		/* Accession of mRNA. */
    char *proteinAcc;		/* Accession of protein. */
    char *geneName;		/* Gene name */
    int locusLinkId;		/* LocusLink id (of mRNA) */
    int size;                   /* mRNA size. */
    int cdsStart, cdsEnd;       /* start/end of coding region (mRNA coords) */
    int ngi;			/* Nucleotide GI number. */
    struct psl *pslList;	/* List of aligments. */
    int geneNameId;		/* Id of gene name in table. */
    int productNameId;		/* Id of product name in table. */
    };

struct hash *hashNextRa(struct lineFile *lf)
/* Read in a record of a .ra file into a hash */
{
struct hash *hash = newHash(5);
char *line, *firstWord;
int lineSize;
int lineCount = 0;

while (lineFileNext(lf, &line, &lineSize))
    {
    if (line[0] == 0)
        break;
    ++lineCount;
    firstWord = nextWord(&line);
    hashAdd(hash, firstWord, cloneString(line));
    }
if (lineCount == 0)
    freeHash(&hash);
return hash;
}

void dotOut()
/* Print dot to standard output and flush it so user can
 * see it right away. */
{
fputc('.', stdout);
fflush(stdout);
}

void parseCds(char *gbCds, int start, int end, int *retStart, int *retEnd)
/* Parse genBank style cds string into something we can use... */
{
char *s = gbCds;
boolean gotStart = FALSE, gotEnd = TRUE;

gotStart = isdigit(s[0]);
s = strchr(s, '.');
if (s == NULL || s[1] != '.')
    errAbort("Malformed GenBank cds %s", gbCds);
s[0] = 0;
s += 2;
gotEnd = isdigit(s[0]);
if (gotStart) start = atoi(gbCds) - 1;
if (gotEnd) end = atoi(s);
*retStart = start;
*retEnd = end;
}

void findCdsStartEndInGenome(struct refSeqInfo *rsi, struct psl *psl, 
	int *retCdsStart, int *retCdsEnd)
/* Convert cdsStart/End from mrna to genomic coordinates. */
{
int strandStart, strandEnd;
int startOffset, endOffset;
int cdsStart, cdsEnd;
int i;

if (psl->strand[0] == '-')
    {
    strandStart = psl->qSize - rsi->cdsEnd;
    strandEnd = psl->qSize - rsi->cdsStart;
    }
else
    {
    strandStart = rsi->cdsStart;
    strandEnd = rsi->cdsEnd;
    }
startOffset = strandStart - psl->qStart;
if (startOffset < 0) startOffset = 0;
endOffset = psl->qEnd - strandEnd;
if (endOffset < 0) endOffset = 0;
#ifdef DEBUG
uglyf("strand %c, size %d, cdsStart %d, strandStart %d, startOffset %d, qStart %d\n",
    psl->strand[0], psl->qSize, rsi->cdsStart, strandStart, startOffset, psl->qStart);
uglyf("    cdsEnd %d, strandEnd %d, endOffset %d, qEnd %d\n",
    rsi->cdsEnd, strandEnd, endOffset, psl->qEnd);
#endif /* DEBUG */

/* Adjust starting pos. */
cdsStart = -1;
for (i=0; i<psl->blockCount; ++i)
    {
    int blockSize = psl->blockSizes[i];
    if (startOffset < blockSize)
	{
        cdsStart = psl->tStarts[i] + startOffset;
	break;
	}
    startOffset -= blockSize;
    }

/* Adjust end pos. */
cdsEnd = -1;
for (i=psl->blockCount-1; i >= 0; --i)
    {
    int blockSize = psl->blockSizes[i];
    if (endOffset < blockSize)
        {
	cdsEnd = psl->tStarts[i] + blockSize - endOffset;
	break;
	}
    endOffset -= blockSize;
    }

#ifdef DEBUG
uglyf("   final cdsStart %d, cdsEnd %d, tStart %d, tEnd %d\n",
	cdsStart, cdsEnd, psl->tStart, psl->tEnd);
#endif /* DEBUG */
if (cdsStart == -1 || cdsEnd == -1)
    {
    cdsEnd = cdsStart = psl->tEnd;
    }
*retCdsStart = cdsStart;
*retCdsEnd = cdsEnd;
}

void processRefSeq(char *raFile, char *pslFile, char *loc2refFile)
/* hgRefSeqMrna - Load refSeq mRNA alignments and other info into 
 * genieKnown table. */
{
struct lineFile *lf;
struct hash *raHash, *rsiHash = newHash(0);
struct refSeqInfo *rsiList = NULL, *rsi;
char *s, *row[5];
int i, dotMod = 0;
int noLocCount = 0;
int rsiCount = 0;
int noProtCount = 0;
struct psl *psl;
int cdsStart, cdsEnd;
char *kgName = "genieKnown";
struct sqlConnection *conn = hgStartUpdate();
struct hash *productHash = loadNameTable(conn, "productName", 16);
struct hash *geneHash = loadNameTable(conn, "geneName", 16);
struct hash *kgHash = newHash(0);
FILE *kgTab = hgCreateTabFile(kgName);
FILE *productTab = hgCreateTabFile("productName");
FILE *geneTab = hgCreateTabFile("geneName");
FILE *knownInfoTab = hgCreateTabFile("knownInfo");
int productNameId, geneNameId;


/* Scan through .ra file and make up start of refSeqInfo
 * objects in hash and list. */
lf = lineFileOpen(raFile, TRUE);
while ((raHash = hashNextRa(lf)) != NULL)
    {
    if (clDots > 0 && ++dotMod == clDots )
        {
	dotMod = 0;
	dotOut();
	}
    AllocVar(rsi);
    slAddHead(&rsiList, rsi);
    if ((s = hashFindVal(raHash, "acc")) == NULL)
        errAbort("No acc near line %d of %s", lf->lineIx, lf->fileName);
    rsi->mrnaAcc = cloneString(s);
    if ((s = hashFindVal(raHash, "siz")) == NULL)
        errAbort("No siz near line %d of %s", lf->lineIx, lf->fileName);
    rsi->size = atoi(s);
    if ((s = hashFindVal(raHash, "gen")) != NULL)
	rsi->geneName = cloneString(s);
    else
        warn("No gene name for %s", rsi->mrnaAcc);
    if ((s = hashFindVal(raHash, "cds")) != NULL)
        parseCds(s, 0, rsi->size, &rsi->cdsStart, &rsi->cdsEnd);
    else
        rsi->cdsEnd = rsi->size;
    if ((s = hashFindVal(raHash, "ngi")) != NULL)
        rsi->ngi = atoi(s);
    rsi->geneNameId = putInNameTable(geneHash, geneTab, rsi->geneName);
    rsi->productNameId = putInNameTable(productHash, productTab, hashFindVal(raHash, "pro"));
    hashAdd(rsiHash, rsi->mrnaAcc, rsi);

    freeHashAndVals(&raHash);
    }
lineFileClose(&lf);
if (clDots) printf("\n");

/* Scan through loc2ref filling in some gaps in rsi. */
lf = lineFileOpen(loc2refFile, TRUE);
while (lineFileRow(lf, row))
    {
    if ((rsi = hashFindVal(rsiHash, row[1])) != NULL)
        {
	rsi->locusLinkId = lineFileNeedNum(lf, row, 0);
	rsi->proteinAcc = cloneString(row[4]);
	}
    }
lineFileClose(&lf);

/* Report how many seem to be missing from loc2ref file. 
 * Write out knownInfo file. */
for (rsi = rsiList; rsi != NULL; rsi = rsi->next)
    {
    ++rsiCount;
    if (rsi->locusLinkId == 0)
        ++noLocCount;
    if (rsi->proteinAcc == NULL)
        ++noProtCount;
    fprintf(knownInfoTab, "%s\t%s\t%s\t%u\t%u\t%s\t%d\t%s\n",
    	rsi->geneName, rsi->mrnaAcc, rsi->geneName,
	rsi->geneNameId, rsi->productNameId,
	(rsi->proteinAcc ? rsi->proteinAcc : ""),
	rsi->ngi, "");
    }
uglyf("Ok 3\n");
if (noLocCount) 
    printf("Missing locusLinkIds for %d of %d\n", noLocCount, rsiCount);
if (noProtCount)
    printf("Missing protein accessions for %d of %d\n", noProtCount, rsiCount);

/* Process alignments and write them out as genes. */
lf = pslFileOpen(pslFile);
dotMod = 0;
while ((psl = pslNext(lf)) != NULL)
    {
    if (clDots > 0 && ++dotMod == clDots )
        {
	dotMod = 0;
	dotOut();
	}
    fprintf(kgTab, "%s\t%s\t%c\t%d\t%d\t",
	psl->qName, psl->tName, psl->strand[0], psl->tStart, psl->tEnd);
    rsi = hashMustFindVal(rsiHash, psl->qName);
    findCdsStartEndInGenome(rsi, psl, &cdsStart, &cdsEnd);
    fprintf(kgTab, "%d\t%d\t", cdsStart, cdsEnd);
    fprintf(kgTab, "%d\t", psl->blockCount);
    for (i=0; i<psl->blockCount; ++i)
	fprintf(kgTab, "%d,", psl->tStarts[i]);
    fprintf(kgTab, "\t");
    for (i=0; i<psl->blockCount; ++i)
	fprintf(kgTab, "%d,", psl->tStarts[i] + psl->blockSizes[i]);
    fprintf(kgTab, "\n");
    }
if (clDots) printf("\n");
uglyf("Loading tables into database\n");
carefulClose(&kgTab);
carefulClose(&productTab);
carefulClose(&geneTab);
carefulClose(&knownInfoTab);
hgLoadTabFile(conn, kgName);
hgLoadTabFile(conn, "productName");
hgLoadTabFile(conn, "geneName");
hgLoadTabFile(conn, "knownInfo");
hgEndUpdate(&conn, "Added known genes");
}

void hgRefSeqMrna(char *database, char *raFile, char *pslFile, 
	char *loc2refFile)
/* hgRefSeqMrna - Load refSeq mRNA alignments and other info into 
 * genieKnown table. */
{
hgSetDb(database);
processRefSeq(raFile, pslFile, loc2refFile);
}

int main(int argc, char *argv[])
/* Process command line. */
{
cgiSpoof(&argc, argv);
clTest = cgiBoolean("test");
clDots = cgiOptionalInt("dots", clDots);
if (argc != 5)
    usage();
hgRefSeqMrna(argv[1], argv[2], argv[3], argv[4]);
return 0;
}
